"""
takes a program spec for execution
"""

from typing import Callable
from pathlib import Path

def chkdct(d: dict, t: dict[str, type]):
    for k in d:
        if k not in t: raise ValueError(f'{k} not in {t.keys()}')
    for k in t:
        if k not in d: raise ValueError(f'{k} not in {d.keys()}')
    for tn,tt in t.items():
        try:
            if not isinstance(d[tn], tt):
                raise ValueError(f'{tn} not of type {tt}')
        except TypeError:
            # some complicated thing not supported by isinstance
            ...


class Rule:
    from .rules import Rule
    @classmethod
    def mk(cls, i: dict) -> Rule:
        assert(isinstance(i, dict))
        if '.py' in i['module']:
            m = Path(i['module'])
            assert(m.exists())
            mod = {}
            _ = exec(open(m).read(), mod)
        else:
            from importlib import import_module
            mod = import_module(i['module']).__dict__
        maker = mod[i['maker']]
        assert(isinstance(maker, Callable))
        params = i['params'] if 'params' in i else {}
        return maker(**params)

#from .engine import Engine
# OO sucks
#class Engine(Engine):  # ❌
class Engine:           # ✔️
    @classmethod
    def args(cls):
        from inspect import signature as sig
        from .engine import Engine
        return {p:v for p,v in sig(Engine.__init__).parameters.items()
            if p not in {'self', }}
        
    @classmethod
    def mk(cls, i: dict):
        assert(isinstance(i, dict))
        args = {k:v.annotation for k,v in cls.args().items()}
        # if                     then don't need these specified
        if i['log']     == False:   args.pop('log_print')
        chkdct(i,  args)

        from .engine import Engine
        if 'log_print' in i:
            if i['log_print']:
                cls.log(enable=True)
        cls.log(enable=False)
        _ = Engine(**i)
        return _
    
    @classmethod
    def log(cls, enable=True, level='INFO'):
        if not enable: return
        from . import logger
        import logging
        logging.basicConfig(force=True) # force removes other loggers that got picked up.
        logger.setLevel(getattr(logging, level))


from dataclasses import dataclass
@dataclass
class Execution:
    params: dict
    rules: list[Rule.Rule]

    from typing import Self
    @classmethod
    def mk(cls, i: dict) -> Self:
        assert(isinstance(i, dict))
        rules = i['rules']
        # /could/ make a case for having a rule /generator/
        # but explicit > implict.
        # i think it's clearer to just list it out
        assert(isinstance(rules, list)) 
        rules = [Rule.mk(r) for r in rules]
        return cls(
            params = i['params'],
            rules = rules
            )
    
    from pyoxigraph import Store
    def engine(self, db: Store):
        _ = {**self.params, 'rules': self.rules, 'db': db}
        _ = Engine.mk(_)
        return _


@dataclass
class Program:
    from pyoxigraph import Store
    db:     Path | Store | None
    execs:  list[Execution]

    @classmethod
    def mkdb(cls, i: Store | Path | str | None) -> Store:
        if i is None:
            return cls.Store()
        elif isinstance(i, Path):
            return cls.Store(i)
        elif isinstance(i, str):
            i = Path(i)
            return cls.Store(i)
        else:
            assert(isinstance(i, cls.Store))
            return i
    
    from typing import Self
    @classmethod
    def mk(cls, i: dict) -> Self:
        # db = i['db'] if 'db' in i else None
        db = cls.mkdb(i['db'])
        execs = [Execution.mk(e) for e in i['execs']]
        return cls(
            db = db,
            execs = execs
        )

    from typing import Self
    @classmethod
    def parse(cls, i: str) -> Self:
        from yaml import safe_load
        _ = safe_load(i)
        _ = cls.mk(_)
        return _

    def run(self) -> Store:
        for e in self.execs:
            e = e.engine(self.db)
            db = e.run()
        return db
        
    __call__ = run





class _reification:
    from typing import Iterable
    from pyoxigraph import Triple, BlankNode
    class terms:
        prefix = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        terms = {'Statement', 'type', 'reifies',
                 'subject', 'predicate', 'object'}
        def __init__(self):
            for t in self.terms:
                setattr(self, t, self.nn(self.prefix, t))
        @staticmethod
        def nn(prefix, term):
            from pyoxigraph import NamedNode
            return NamedNode(f"{prefix}{term}")
    terms = terms()
    def standard(self, str: Iterable[Triple], deterministic_nested=True) -> Iterable[Triple]:
        from .canon import quads
        uri = quads.deanon.defaults.uri
        T = self.Triple
        trm = self.terms
        old2new = {}
        str = frozenset(str)  # need two passes for deterministic_nested case

        def isnested(t):
            # nested/reifying  triple https://www.w3.org/TR/rdf12-concepts/#dfn-reifying-triple
            return (t.predicate == trm.reifies) and isinstance(t.object, T)
        for t in str:
            if isnested(t):
                ndt = t.object                            # always  here
                assert(not isinstance(t.subject, T))      # but not here
                # https://github.com/oxigraph/oxigraph/issues/1286
                # <<d:s d:p d:o>> m:p m:o .
                # <<d:s d:p d:o>> m:p2 m:o2 .
                # # not the same as
                # <<d:s d:p d:o>> m:p m:o; m:p2 m:o2 .
                # doesn't make sense for rdf-engine
                # deterministic. reduce blank nodes in output
                id = t.subject
                if deterministic_nested:
                    old = id
                    id = hash(ndt)
                    id = abs(id)
                    id = self.terms.nn(uri, id)
                    new = id
                else:
                    id = t.subject
                    old = new = id
                old2new[old] = new
                    
                yield T(id, trm.type,       trm.Statement)
                yield T(id, trm.subject,    ndt.subject)
                yield T(id, trm.predicate,  ndt.predicate)
                yield T(id, trm.object,     ndt.object)
                # meta       t.reifies
                #yield T(id, t.predicate,    t.object)
                # meta in below loop

        for t in str:
            if not isnested(t):
                if t.subject in old2new:
                    t = T(old2new[t.subject], t.predicate, t.object)
                if t.object in old2new:
                    t = T(t.subject, t.predicate, old2new[t.object])
                yield t
                
    # :man :hasSpouse :woman .
    # :id1 rdf:type rdf:Statement ;
        # rdf:subject :man ;
        # rdf:predicate :hasSpouse ;
        # rdf:object :woman ;
        # :startDate "2020-02-11"^^xsd:date .
    # <<:man :hasSpouse :woman>> :startDate "2020-02-11"^^xsd:date .
    class query:
        # https://www.ontotext.com/knowledgehub/fundamentals/what-is-rdf-star/
        class parts:
            rei = """
            ?reification a rdf:Statement .
            ?reification rdf:subject ?subject .
            ?reification rdf:predicate ?predicate .
            ?reification rdf:object ?object .
            """
            left =  "?reification ?p ?o."
            right = "?s ?p ?reification."
            reil = rei + left
            reir = rei + right
            union = f"""
            {rei}
            {{{left}}} union {{{right}}}
            """
        std2str = """
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        CONSTRUCT {
            #<<?subject ?predicate ?object>> ?p ?o .
            #?s ?p <<?subject ?predicate ?object>> .
            ?reification rdf:reifies <<(?subject ?predicate ?object)>>. # tricky!!
            side
        } WHERE {
            part
            FILTER (?p NOT IN (rdf:subject, rdf:predicate, rdf:object) &&
            (?p != rdf:type && ?object != rdf:Statement))
        }
        """
        std2strl = std2str.replace('part', parts.reil).replace('side', parts.left)
        std2strr = std2str.replace('part', parts.reir).replace('side', parts.right)
        delrei = """
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        DELETE where  { part }
        """
        delreil = delrei.replace('part', parts.reil)
        delreir = delrei.replace('part', parts.reir)
    def star(self, std: Iterable[Triple],) -> Iterable[Triple]:
        from pyoxigraph import Store, Quad
        s = Store()
        s.bulk_extend(Quad(*t) for t in std)
        yield from s.query( self.query.std2strl)
        yield from s.query( self.query.std2strr)
        s.update(           self.query.delreil)
        s.update(           self.query.delreir)
        #                             the 'regular' ones
        yield from s.query("construct {?s ?p ?o} where {?s ?p ?o.}")


reification = _reification()


from .db import Ingestable
from typing import Iterable
from pyoxigraph import Quad
def quads(i: Ingestable) -> Iterable[Quad]:
    from .db import ingest, Store
    yield from ingest(Store(), i)
